import json
import logging
import os.path
import shlex
import subprocess
import time
import urllib


class BwProcess:
  def __init__(self, bw_cli_path):
    if not os.access(bw_cli_path, os.X_OK):
      raise ValueError(f'{bw_cli_path} does not exists or is not executable')
    self.__bw = bw_cli_path
    self.__bw_process = None

  def run(self, *args, parse_output=False, timeout=30):
    logging.info('Running %s', shlex.join(args))
    args = [self.__bw] + list(args)
    output = subprocess.check_output(args, timeout=timeout)
    if parse_output:
      return json.loads(output)

  def start_serve(self, username, password, host, port):
    status_output = self.run('status', parse_output=True)
    assert status_output['status'] == 'unauthenticated', f'Unexpected status: {status_output}'
    self.run('login', username, password)

    self.__bw_process = subprocess.Popen(
      [self.__bw, 'serve', '--hostname', host, '--port', str(port)],
      bufsize=0, stderr=subprocess.STDOUT, stdout=subprocess.PIPE, stdin=subprocess.PIPE,
    )
    logging.info('Started bw serve with pid %d', self.__bw_process.pid)

    for _ in range(0, 300):
      time.sleep(0.1)
      logging.info('Checking if bw serve is up')
      if self.__bw_process.poll() is not None:
        self.terminate_serve()
        outs, errs = self.__bw_process.communicate(timeout=15)
        raise AssertionError(
          f'bw serve (PID: {self.__bw_process.pid}) unexpectedly reports as terminated'
          f'exit code: {self.__bw_process.returncode}, stdout: {outs}, stderr: {errs}')
      try:
        urllib.request.urlopen(f'http://{host}:{port}/', timeout=1)
      except urllib.error.HTTPError:
        # url responds with a 404
        logging.info('bw serve up')
        break
      except urllib.error.URLError:
        pass
    else:
      self.terminate_serve()
      outs, errs = self.__bw_process.communicate(timeout=15)
      raise AssertionError(
        f'bw serve failed to start (PID: {self.__bw_process.pid}). '
        f'exit code: {self.__bw_process.returncode}, stdout: {outs}, stderr: {errs}')

  def terminate_serve(self):
    logging.info('Terminating bw serve')
    self.__bw_process.terminate()
    try:
      self.__bw_process.communicate(timeout=10)
      logging.info('bw serve (PID: %d) terminated', self.__bw_process.pid)
    except subprocess.TimeoutExpired:
      logging.warn('bw serve didn\'t terminate, killing it')
      self.__bw_process.kill()
    self.__bw_process = None
    self.run('logout')




from typing import Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="LockunlockSuccessData")


@_attrs_define
class LockunlockSuccessData:
    """
    Attributes:
        message (Union[Unset, str]):
        no_color (Union[Unset, bool]):
        object_ (Union[Unset, str]):
        title (Union[Unset, str]):
    """

    message: Union[Unset, str] = UNSET
    no_color: Union[Unset, bool] = UNSET
    object_: Union[Unset, str] = UNSET
    title: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        message = self.message

        no_color = self.no_color

        object_ = self.object_

        title = self.title

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if message is not UNSET:
            field_dict["message"] = message
        if no_color is not UNSET:
            field_dict["noColor"] = no_color
        if object_ is not UNSET:
            field_dict["object"] = object_
        if title is not UNSET:
            field_dict["title"] = title

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        message = d.pop("message", UNSET)

        no_color = d.pop("noColor", UNSET)

        object_ = d.pop("object", UNSET)

        title = d.pop("title", UNSET)

        lockunlock_success_data = cls(
            message=message,
            no_color=no_color,
            object_=object_,
            title=title,
        )

        lockunlock_success_data.additional_properties = d
        return lockunlock_success_data

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties
